import logging
import os
import pprint
import re
import shutil
import sys
import subprocess
import tempfile
import traceback
import time
import zipfile
from datetime import datetime
from decimal import Decimal

import praw
import yaml
from jinja2 import Environment, PackageLoader, StrictUndefined
from sqlalchemy import create_engine
from prawcore.exceptions import PrawcoreException, ResponseException

from . import actions, stats
from .coin import Coin
from .const import __version__
from .user import User
from .util import log_function

logger = logging.getLogger(__package__)
logger.setLevel(logging.DEBUG)
log_decorater = log_function(klass="NyanTip", log_method=logger.info)

EXCEPTION_SLEEP_TIME = 60  # seconds


class NyanTip:
    CONFIG_NAME = "nyantip.yml"
    PERIODIC_TASKS = {
        "expire_pending_tips": {"period": 60},
        "load_banned_users": {"period": 300},
        "update_statistics": {"period": 900},
    }

    def __init__(self):
        self._running = False
        self.banned_users = None
        self.bot = None
        self.commands = []
        self.config = self.parse_config()
        self.database = None
        self.exception_user = None
        self.reddit = None
        self.templates = Environment(
            loader=PackageLoader(__package__),
            trim_blocks=True,
            undefined=StrictUndefined,
        )

        self.coin = Coin(config=self.config["coin"])

    @staticmethod
    def config_path():
        if "APPDATA" in os.environ:  # Windows
            os_config_path = os.environ["APPDATA"]
        elif "XDG_CONFIG_HOME" in os.environ:  # Modern Linux
            os_config_path = os.environ["XDG_CONFIG_HOME"]
        elif "HOME" in os.environ:  # Legacy Linux
            os_config_path = os.path.join(os.environ["HOME"], ".config")
        else:
            raise Exception(
                "APPDATA, XDG_CONFIG_HOME, nor HOME environment variables are set"
            )
        path = os.path.join(os_config_path, NyanTip.CONFIG_NAME)
        if not os.path.isfile(path):
            raise Exception(f"Config file does not exist at {path}")
        return path

    @staticmethod
    def config_to_decimal(container, key):
        assert isinstance(container[key], str)
        container[key] = Decimal(container[key]).normalize()

    @classmethod
    def parse_config(cls):
        path = cls.config_path()
        logger.debug(f"reading config from {path}")
        with open(path) as fp:
            config = yaml.safe_load(fp)
        cls.config_to_decimal(config["coin"], "minimum_tip")
        cls.config_to_decimal(config["coin"], "minimum_withdraw")
        cls.config_to_decimal(config["coin"], "transaction_fee")
        return config

    def _run_loop(self):
        for item in self.reddit.inbox.stream(pause_after=4):
            if item is None:
                now = time.time()
                for task_name, task_metadata in self.PERIODIC_TASKS.items():
                    if now >= task_metadata.setdefault(
                        "next_run_time", now + task_metadata["period"]
                    ):
                        getattr(self, task_name)()
                        now = time.time()
                        task_metadata["next_run_time"] = now + task_metadata["period"]
            else:
                try:
                    self.process_message(item)
                except Exception:
                    item_info = pprint.pformat(vars(item), indent=4)
                    logger.exception(
                        f"Exception processing the following item:\n{item_info}"
                    )

                    if self.exception_user:
                        message = f"Exception\n{traceback.format_exc()}\nItem:\n{item_info}".replace(
                            "\n", "\n\n"
                        )
                        self.exception_user.message(
                            message=message, subject="nyantip exception"
                        )

                    time.sleep(60)  # Let's slow things down if there are issues
                    continue
                item.mark_read()

    def backup(self):
        backup_name = f"backup_nyantip_{datetime.now().strftime('%Y%m%d%H%M')}"
        backup_passphrase = self.config["backup_passphrase"]

        with tempfile.NamedTemporaryFile() as temp_fp:
            with zipfile.ZipFile(temp_fp, mode="w") as zip_fp:
                # Backup config
                zip_fp.write(
                    self.config_path(), arcname=f"{backup_name}/{self.CONFIG_NAME}"
                )

                # Backup database
                database = self.config["database"]["name"]
                with tempfile.NamedTemporaryFile() as database_fp:
                    mysqldump_command = [
                        "mysqldump",
                        "--databases",
                        database,
                        "--host",
                        self.config["database"]["host"],
                        "--port",
                        str(self.config["database"]["port"]),
                        "--single-transaction",
                    ]
                    password = self.config["database"]["password"]
                    if password:
                        mysqldump_command.extend(["--password", password])
                    user = self.config["database"]["user"]
                    if user:
                        mysqldump_command.extend(["--user", user])
                    subprocess.run(mysqldump_command, check=True, stdout=database_fp)
                    zip_fp.write(
                        database_fp.name, arcname=f"{backup_name}/{database}.sql"
                    )

                # Backup wallet
                prefix = self.config["coin"]["name"].lower()
                with tempfile.NamedTemporaryFile() as wallet_fp:
                    self.coin.connection.backupwallet(wallet_fp.name)
                    zip_fp.write(
                        wallet_fp.name, arcname=f"{backup_name}/{prefix}_wallet.dat"
                    )

            if backup_passphrase:
                import gnupg

                gpg = gnupg.GPG()
                temp_fp.seek(0)
                gpg.encrypt_file(
                    temp_fp,
                    output=f"{backup_name}.zip.gpg",
                    passphrase=backup_passphrase,
                    recipients=None,
                    symmetric="AES256",
                )
            else:
                shutil.copy(temp_fp.name, f"{backup_name}.zip")

    def connect_to_database(self):
        info = self.config["database"]
        name = info["name"]
        user = self.config["database"]["user"]
        logger.info(f"Connecting to database {name} as {user or 'anonymous'}")

        credentials = f"{user}:{self.config['database']['password']}@" if user else ""
        self.database = create_engine(
            f"mysql+mysqldb://{credentials}{info['host']}:{info['port']}/{name}?charset=utf8mb4"
        )

    def connect_to_reddit(self):
        self.reddit = praw.Reddit(
            check_for_updates=False,
            ratelimit_seconds=600,
            user_agent=f"nyantip/{__version__} by u/bboe",
            **self.config["reddit"],
        )
        try:
            self.reddit.user.me()  # Ensure credentials are correct
        except ResponseException as exception:
            if exception.response.status_code == 401:
                logger.error("Invalid reddit credentials")
                sys.exit(1)
            raise
        if self.config["exception_user"]:
            self.exception_user = self.reddit.redditor(self.config["exception_user"])

    @log_decorater
    def expire_pending_tips(self):
        pending_hours = int(self.config["pending_hours"])

        for action in actions.actions(
            action="tip",
            created_at=f"created_at < DATE_SUB(NOW(), INTERVAL {pending_hours} HOUR)",
            nyantip=self,
            status="pending",
        ):
            action.expire()

    def load_banned_users(self):
        self.banned_users = set()
        for username in self.config.get("banned", []):
            self.banned_users.add(self.reddit.redditor(username))

        subreddit = self.config["reddit"]["subreddit"]
        for user in self.reddit.subreddit(subreddit).banned(limit=None):
            self.banned_users.add(user)

        logger.info(f"Loaded {len(self.banned_users)} banned user(s)")

    def no_match(self, *, message, message_type):
        logger.info("no match")
        response = self.templates.get_template("didnt-understand.tpl").render(
            config=self.config,
            message=message,
            message_type=message_type,
        )
        User(name=message.author.name, nyantip=self, redditor=message.author).message(
            body=response,
            message=message,
            subject="What?",
        )

    def prepare_commands(self):
        for action, action_config in self.config["commands"].items():
            if isinstance(action_config, str):
                expression = action_config
                command = {
                    "action": action,
                    "only": "message",
                    "regex": re.compile(action_config, re.IGNORECASE | re.DOTALL),
                }
                command["regex"] = re.compile(expression, re.IGNORECASE | re.DOTALL)
                logger.debug(f"ADDED REGEX for {action}: {command['regex'].pattern}")
                self.commands.append(command)
                continue

            for _, option in sorted(action_config.items()):
                expression = (
                    option["regex"]
                    .replace("{REGEX_ADDRESS}", self.coin.config["regex"])
                    .replace("{REGEX_AMOUNT}", r"(\d{1,9}(?:\.\d{0,8})?)")
                    .replace(
                        "{REGEX_KEYWORD}", f"({'|'.join(self.config['keywords'])})"
                    )
                    .replace("{REGEX_USERNAME}", r"/?u/([\w-]{3,20})")
                    .replace("{BOT_NAME}", f"/?u/{self.config['reddit']['username']}")
                )

                command = {
                    "action": action,
                    "address": option["address"],
                    "amount": option["amount"],
                    "destination": option["destination"],
                    "keyword": option["keyword"],
                    "only": option.get("only"),
                }

                command["regex"] = re.compile(expression, re.IGNORECASE | re.MULTILINE)
                logger.debug(f"ADDED REGEX for {action}: {command['regex'].pattern}")
                self.commands.append(command)

    def process_message(self, message):
        message_type = "comment" if message.was_comment else "message"
        if not message.author:
            logger.info(f"ignoring {message_type} with no author")
            return

        if actions.check_action(
            message_id=message.id,
            nyantip=self,
        ):
            logger.warning(
                "duplicate action detected (message.id %s), ignoring",
                message.id,
            )
            return
        if message.author == self.config["reddit"]["username"]:
            logger.debug("ignoring message from self")
            return
        if message.author in self.banned_users:
            logger.info(f"ignoring message from banned user {message.author}")
            return

        for command in self.commands:
            match = command["regex"].search(message.body)
            if match:
                action = command["action"]
                if command["only"] and message_type != command["only"]:
                    logger.debug(
                        f"ignoring {action} because it's only permitted in {command['only']}"
                    )
                    continue
                break
        else:
            logger.debug("no match found")
            self.no_match(message=message, message_type=message_type)
            return

        address = match.group(command["address"]) if command.get("address") else None
        amount = match.group(command["amount"]) if command.get("amount") else None
        destination = (
            match.group(command["destination"]) if command.get("destination") else None
        )
        keyword = match.group(command["keyword"]) if command.get("keyword") else None

        assert not (address and destination)  # Both should never be set
        if not address and not destination:
            if message.was_comment:
                destination = message.parent().author.name
                assert destination

        logger.info(f"{action} from {message.author} ({message_type} {message.id})")
        logger.debug(f"message body:\n<begin>\n{message.body}\n</end>")
        actions.Action(
            action=action,
            amount=amount,
            destination=address or destination,
            keyword=keyword,
            message=message,
            nyantip=self,
        ).perform()

    def run(self):
        self.bot = User(name=self.config["reddit"]["username"], nyantip=self)
        self.prepare_commands()
        self.connect_to_database()
        self.connect_to_reddit()
        self.run_self_check()

        # Run these tasks every start up
        self.load_banned_users()
        self.expire_pending_tips()

        logger.info(f"Bot starting v{__version__}")
        self._running = True
        while self._running:
            try:
                self._run_loop()
            except KeyboardInterrupt:
                self._running = False
            except PrawcoreException:
                logger.exception(
                    f"PrawcoreException in runloop. Sleeping for {EXCEPTION_SLEEP_TIME} seconds."
                )
                time.sleep(EXCEPTION_SLEEP_TIME)
        logger.info(f"Bot stopped gracefully v{__version__}")

    @log_decorater
    def run_self_check(self):
        # Ensure bot is a registered user
        if not self.bot.is_registered():
            self.bot.register()

        # Ensure coin balance is positive
        balance = self.coin.connection.getbalance()
        if balance < 0:
            raise Exception(f"negative wallet balance: {balance}")

        # Ensure pending tips <= bot's escrow balance
        balance = self.bot.balance(kind="tip")
        pending_tips = sum(
            x.amount
            for x in actions.actions(action="tip", nyantip=self, status="pending")
        )
        if balance < pending_tips:
            raise Exception(
                f"Bot's escrow balance ({balance}) < total pending tips ({pending_tips})"
            )

        # Ensure user account balances are not negative
        for row in self.database.execute(
            "SELECT username FROM users ORDER BY username"
        ):
            username = row["username"]
            if User(name=username, nyantip=self).balance(kind="tip") < 0:
                raise Exception(f"{username} has a negative balance")

    def update_statistics(self):
        stats.update_stats(nyantip=self)
        stats.update_tips(nyantip=self)



import logging
import time

logger = logging.getLogger(__package__)


class DummyMessage:
    def __init__(self, author, context):
        self.author = author
        self.context = context


def log_function(*fields, klass=None, log_method=None, log_response=False):
    if not log_method:
        log_method = logger.debug

    def decorator(function):
        def wrapped(*args, **kwargs):
            arguments = ",".join(
                f"{field}={kwargs[field]!r}" for field in fields if kwargs.get(field)
            )

            start = time.time() * 1000
            response = function(*args, **kwargs)
            duration = time.time() * 1000 - start

            description = f"{function.__name__}({arguments})"
            if klass:
                description = f"{klass}.{description}"
            if log_response:
                description = f"{description} = {response!r}"

            log_method(f"{description} in {duration:0.4f} ms")
            return response

        return wrapped

    return decorator



import argparse
import logging

from .bot import NyanTip
from .const import __version__  # noqa

logging.basicConfig(
    datefmt="%H:%M:%S",
    format="%(asctime)s %(levelname)-8s %(name)-12s %(message)s",
)
logging.getLogger("BitcoinRPC").setLevel(logging.DEBUG)


def main():
    parser = argparse.ArgumentParser(
        description="Run nyantip bot or associated utilities"
    )
    subparsers = parser.add_subparsers(dest="command", metavar="", title="subcommands")
    subparsers.add_parser("backup", help="Backup config, database, and wallet")

    arguments = parser.parse_args()
    if arguments.command == "backup":
        NyanTip().backup()
    else:
        NyanTip().run()



"""
    This file is part of ALTcointip.

    ALTcointip is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    ALTcointip is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ALTcointip.  If not, see <http://www.gnu.org/licenses/>.
"""
import configparser
import logging
import os
import sys

from .rpc import Rpc
from .util import log_function

logger = logging.getLogger(__package__)


class Coin:
    def __init__(self, config):
        self.config = config
        rpc_config = read_coin_config(config["config_file"])

        logger.debug(f"connecting to {config['name']}")
        self.connection = Rpc(
            f"http://{rpc_config['rpcuser']}:{rpc_config['rpcpassword']}@127.0.0.1:{rpc_config['rpcport']}"
        )

        logger.info(f"Setting transaction fee of {config['transaction_fee']}")
        try:
            self.connection.settxfee(config["transaction_fee"])
        except ConnectionRefusedError:
            logger.error(
                f"error connecting to {config['name']} ({config['config_file']}) is it running?"
            )
            sys.exit(1)

    def __str__(self):
        return self.config["name"]

    def balance(self, *, minconf, user):
        return self.connection.getbalance(user, minconf).normalize()

    def generate_address(self, *, user):
        passphrase = self.config.get("walletpassphrase")

        if passphrase:
            self.connection.walletpassphrase(passphrase, 1)

        try:
            return self.connection.getnewaddress(user)
        finally:
            if passphrase:
                self.connection.walletlock()

    @log_function("amount", "destination", "source", klass="Coin")
    def send(self, *, amount, destination, source):
        self.connection.move(source.name, destination.name, amount)

    @log_function("amount", "address", "source", klass="Coin", log_response=True)
    def transfer(self, *, address, amount, source):
        passphrase = self.config.get("walletpassphrase")

        if passphrase:
            self.connection.walletpassphrase(passphrase, 1)

        try:
            return self.connection.sendfrom(
                source, address, amount, self.config["minconf"]["withdraw"]
            )
        finally:
            if passphrase:
                self.connection.walletlock()

    @log_function("address", klass="Coin", log_response=True)
    def validate(self, *, address):
        return self.connection.validateaddress(address).get("isvalid", False)


def read_coin_config(filename):
    config = configparser.ConfigParser()
    with open(os.path.expanduser(filename)) as fp:
        config.read_string(f"[DEFAULT]\n{fp.read()}")
    return dict(config.items("DEFAULT"))



"""
    This file is part of ALTcointip.

    ALTcointip is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    ALTcointip is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ALTcointip.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from decimal import Decimal
from functools import partial

from praw.models import Comment

from . import stats, user
from .util import DummyMessage

logger = logging.getLogger(__package__)


class Action(object):
    def __init__(
        self,
        *,
        action,
        nyantip,
        message,
        amount=None,
        destination=None,
        keyword=None,
    ):

        self.action = action
        self.amount = amount
        self.message = message
        self.nyantip = nyantip
        self.source = user.User(
            nyantip=nyantip, name=message.author.name, redditor=message.author
        )
        self.transaction_id = None

        if action == "tip":
            self.destination = user.User(name=destination, nyantip=nyantip)
        else:
            self.destination = destination

        if self.action in ["tip", "withdraw"]:
            if keyword:
                assert self.amount is None
                value = self.nyantip.config["keywords"][keyword.lower()]
                if isinstance(value, Decimal):
                    self.amount = value
                else:
                    assert isinstance(value, str)
                    logger.debug(f"__init__(): evaluating {value!r}")
                    self.amount = eval(value)
            elif isinstance(amount, str):
                assert self.amount.replace(".", "").isnumeric()
                self.amount = Decimal(self.amount)

            assert isinstance(self.amount, Decimal)

    def __str__(self):
        return f"<Action: action={self.action}, amount={self.amount} destination={self.destination} source={self.message.author}>"

    @property
    def _amount_formatted(self):
        return self._format_coin(self.amount)

    def _fail(self, subject, template, save=True, **template_args):
        response = self.nyantip.templates.get_template(template).render(
            config=self.nyantip.config, message=self.message, **template_args
        )
        self.source.message(body=response, message=self.message, subject=subject)
        if save:
            self.save(status="failed")
        return False

    def _format_coin(self, quantity):
        return f"{quantity:f} {self.nyantip.config['coin']['name']}"

    def _safe_send(self, *, amount=None, destination, on_success, source):
        if amount is None:
            amount = self.amount

        try:
            self.nyantip.coin.send(
                amount=amount,
                destination=destination,
                source=source,
            )
        except Exception:
            logger.exception(f"action_{self.action}(): failed")
            return self._fail(
                f"{self.action} failed",
                "tip-went-wrong.tpl",
                action_name=self.action,
                amount_formatted=self._amount_formatted,
                destination=None,
                to_address=False,
            )
        try:
            on_success()
        except Exception:
            # If the callback fails, we need to undo the send
            logger.warning("rolling back the previous transfer")
            self.nyantip.coin.send(
                amount=amount,
                destination=source,
                source=destination,
            )
            raise
        return True

    def action_accept(self):
        pending_actions = actions(
            action="tip",
            nyantip=self.nyantip,
            status="pending",
            destination=self.source,
        )
        if not pending_actions:
            return self._fail("accept failed", "no-pending-tips.tpl")

        if not self.source.is_registered():
            self.source.register()

        users_to_update = set()
        for action in pending_actions:
            if not self._safe_send(
                amount=action.amount,
                destination=self.source,
                on_success=partial(action.save, status="completed"),
                source=self.nyantip.bot,
            ):
                self.save(status="failed")
                return
            users_to_update.add(action.source.name)

            response = self.nyantip.templates.get_template("confirmation.tpl").render(
                amount_formatted=action._amount_formatted,
                config=self.nyantip.config,
                destination=action.destination,
                message=action.message,
                title="verified",
                to_address=False,
                transaction_id=None,
            )
            action.source.message(
                body=response,
                message=action.message,
                reply_to_comment=True,
                subject="tip succeeded",
            )

        self.save(status="completed")
        self.action = "info"
        self.action_info(save=False)

        stats.update_user_stats(nyantip=self.nyantip, username=self.source.name)
        for username in users_to_update:
            stats.update_user_stats(nyantip=self.nyantip, username=username)

    def action_decline(self):
        pending_actions = actions(
            action="tip",
            nyantip=self.nyantip,
            status="pending",
            destination=self.source,
        )
        if not pending_actions:
            return self._fail("decline failed", "no-pending-tips.tpl")

        for action in pending_actions:
            if not self._safe_send(
                amount=action.amount,
                destination=action.source,
                on_success=partial(action.save, status="declined"),
                source=self.nyantip.bot,
            ):
                self.save(status="failed")
                return

            response = self.nyantip.templates.get_template("confirmation.tpl").render(
                amount_formatted=action._amount_formatted,
                config=self.nyantip.config,
                destination=action.destination,
                message=action.message,
                title="declined",
                to_address=False,
                transaction_id=None,
            )
            action.source.message(
                body=response, message=action.message, subject="tip declined"
            )

        self.save(status="completed")
        response = self.nyantip.templates.get_template(
            "pending-tips-declined.tpl"
        ).render(
            config=self.nyantip.config,
            message=self.message,
        )
        self.source.message(
            body=response, message=self.message, subject="declined succeeded"
        )

    def action_history(self):
        history = []

        response = self.nyantip.database.execute(
            self.nyantip.config["sql"]["history"], (self.source, self.source)
        )
        for row in response:
            history_entry = []
            for key in response.keys():
                history_entry.append(
                    stats.format_value(
                        config=self.nyantip.config,
                        compact=True,
                        key=key,
                        username=self.source.name,
                        value=row[key],
                    )
                )
            history.append(history_entry)

        if history:
            response = self.nyantip.templates.get_template("history.tpl").render(
                config=self.nyantip.config,
                history=history,
                keys=response.keys(),
                message=self.message,
            )
        else:
            response = self.nyantip.templates.get_template("history-empty.tpl").render(
                config=self.nyantip.config,
                message=self.message,
            )
        self.message.reply(response)
        self.save(status="completed")

    def action_info(self, save=True):
        if not self.source.is_registered():
            return self._fail("info failed", "not-registered.tpl", save=save)

        balance = self.nyantip.coin.balance(
            minconf=self.nyantip.coin.config["minconf"]["tip"],
            user=self.source.name,
        )
        address = self.nyantip.database.execute(
            "SELECT address FROM users WHERE username = %s", self.source
        ).scalar_one()

        response = self.nyantip.templates.get_template("info.tpl").render(
            action=self,
            address=address,
            balance=balance,
            config=self.nyantip.config,
            message=self.message,
        )
        self.message.reply(response)

        if save:
            self.save(status="completed")

        return True

    def action_register(self):
        if self.source.is_registered():
            logger.debug(
                f"register({self.source}): user already exists; ignoring request"
            )
            self.save(status="failed")
        else:
            self.source.register()
            self.save(status="completed")
        self.action = "info"
        self.action_info(save=False)

    def action_tip(self):
        assert self.destination
        if not self.validate():
            return

        if not self._safe_send(
            destination=self.destination,
            on_success=partial(self.save, status="completed"),
            source=self.source,
        ):
            return

        response = self.nyantip.templates.get_template("confirmation.tpl").render(
            amount_formatted=self._amount_formatted,
            config=self.nyantip.config,
            destination=self.destination,
            message=self.message,
            title="verified",
            to_address=False,
            transaction_id=None,
        )
        self.source.message(
            body=response,
            message=self.message,
            reply_to_comment=True,
            subject="tip succeeded",
        )

        dummy_message = DummyMessage(self.destination, self.message.context)
        response = self.nyantip.templates.get_template("tip-received.tpl").render(
            amount_formatted=self._amount_formatted,
            config=self.nyantip.config,
            dummy_message=dummy_message,
            source=self.message.author,
        )
        self.destination.message(body=response, subject="tip received")

        stats.update_user_stats(nyantip=self.nyantip, username=self.source.name)
        stats.update_user_stats(nyantip=self.nyantip, username=self.destination.name)

    def action_withdraw(self):
        assert self.destination
        if not self.validate():
            return

        try:
            self.transaction_id = self.nyantip.coin.transfer(
                address=self.destination, amount=self.amount, source=self.source.name
            )
        except Exception:
            logger.exception("action_withdraw(): failed")
            return self._fail(
                "withdraw failed",
                "tip-went-wrong.tpl",
                action_name=self.action,
                amount_formatted=self._amount_formatted,
                destination=self.destination,
                to_address=True,
            )

        self.save(status="completed")
        response = self.nyantip.templates.get_template("confirmation.tpl").render(
            amount_formatted=self._amount_formatted,
            config=self.nyantip.config,
            destination=self.destination,
            message=self.message,
            title="verified",
            to_address=True,
            transaction_id=self.transaction_id,
        )
        self.source.message(
            body=response, message=self.message, subject="withdraw succeeded"
        )

    def expire(self):
        if not self._safe_send(
            destination=self.source,
            on_success=partial(self.save, status="expired"),
            source=self.nyantip.bot,
        ):
            return

        response = self.nyantip.templates.get_template("confirmation.tpl").render(
            amount_formatted=self._amount_formatted,
            config=self.nyantip.config,
            destination=self.destination,
            message=self.message,
            title="expired",
            to_address=False,
            transaction_id=None,
        )
        self.source.message(body=response, message=self.message, subject="tip expired")

    def perform(self):
        if self.action == "accept":
            self.action_accept()
        elif self.action == "decline":
            self.action_decline()
        elif self.action == "history":
            self.action_history()
        elif self.action == "info":
            self.action_info()
        elif self.action == "register":
            self.action_register()
        elif self.action == "tip":
            self.action_tip()
        else:
            assert self.action == "withdraw"
            self.action_withdraw()

    def save(self, *, status):
        permalink = None
        if isinstance(self.message, Comment):
            if "context" in self.message.__dict__ and self.message.context:
                permalink = self.message.context
            else:
                if "permalink" not in self.message.__dict__:
                    logging.warning("This refresh shouldn't be necessary")
                    logging.warning(vars(self.message))
                    self.message.refresh()
                permalink = f"{self.message.permalink}?context=3"

        result = self.nyantip.database.execute(
            "REPLACE INTO actions (action, amount, destination, message_id, message_timestamp, path, source, status, transaction_id) VALUES (%s, %s, %s, %s, FROM_UNIXTIME(%s), %s, %s, %s, %s)",
            (
                self.action,
                self.amount,
                self.destination,
                self.message.id,
                self.message.created_utc,
                permalink,
                self.source.name,
                status,
                self.transaction_id,
            ),
        )
        assert 1 <= result.rowcount <= 2

    def validate(self):
        subject = f"{self.action} failed"

        # First see if the author is registered
        if not self.source.is_registered():
            return self._fail(subject, "not-registered.tpl")

        # Second ensure the amount is larger than the necessary minimum
        minimum = self.nyantip.coin.config[f"minimum_{self.action}"]
        if self.amount < minimum:
            return self._fail(
                subject,
                "tip-below-minimum.tpl",
                action_name=self.action,
                amount_formatted=self._amount_formatted,
                minimum_formatted=self._format_coin(minimum),
            )

        # Then verify they have sufficient balance
        balance = self.source.balance(kind=self.action)
        balance_needed = self.amount
        if self.action == "withdraw":
            balance_needed += self.nyantip.coin.config["transaction_fee"]
        if balance < balance_needed:
            return self._fail(
                subject,
                "tip-low-balance.tpl",
                action_name=self.action,
                balance_formatted=self._format_coin(balance),
                balance_needed_formatted=self._format_coin(balance_needed),
            )

        if self.action == "tip":
            if self.source == self.destination:
                return self._fail(subject, "cant-send.tpl", destination="yourself")

            if self.nyantip.bot == self.destination:
                return self._fail(subject, "cant-send.tpl", destination="me")

            if not self.destination.is_redditor():
                return self._fail(
                    subject, "not-on-reddit.tpl", destination=self.destination
                )

            if check_action(
                action="tip",
                destination=self.destination,
                nyantip=self.nyantip,
                source=self.source,
                status="pending",
            ):
                return self._fail(
                    subject, "tip-already-pending.tpl", destination=self.destination
                )

            if not self.destination.is_registered():
                # Perform a pending transfer to escrow
                self.nyantip.coin.send(
                    amount=self.amount,
                    destination=self.nyantip.bot,
                    source=self.source,
                )
                self.save(status="pending")

                response = self.nyantip.templates.get_template(
                    "confirmation.tpl"
                ).render(
                    amount_formatted=self._amount_formatted,
                    config=self.nyantip.config,
                    destination=self.destination,
                    message=self.message,
                    title="pending accept",
                    to_address=False,
                    transaction_id=None,
                )
                self.source.message(
                    body=response,
                    message=self.message,
                    reply_to_comment=True,
                    subject="tip pending accept",
                )

                dummy_message = DummyMessage(self.destination, self.message.context)
                response = self.nyantip.templates.get_template(
                    "tip-pending.tpl"
                ).render(
                    amount_formatted=self._amount_formatted,
                    config=self.nyantip.config,
                    dummy_message=dummy_message,
                    source=self.message.author,
                )
                self.destination.message(body=response, subject="tip pending")
                return False
        elif not self.nyantip.coin.validate(address=self.destination):
            return self._fail(
                subject, "address-invalid.tpl", destination=self.destination
            )
        return True


def actions(
    *,
    action=None,
    created_at=None,
    destination=None,
    message_id=None,
    nyantip=None,
    source=None,
    status=None,
    _check=False,
):
    arguments = []
    filters = []
    for attribute in ("action", "message_id", "destination", "source", "status"):
        value = locals()[attribute]
        if value:
            arguments.append(value)
            filters.append(f"{attribute} = %s")

    if created_at:
        filters.append(created_at)

    sql_where = f" WHERE {' AND '.join(filters)}"
    sql = f"SELECT * FROM actions{sql_where}"

    logger.debug(f"query: {sql} {arguments}")
    response = nyantip.database.execute(sql, arguments)

    if _check:
        return response.rowcount > 0

    if response.rowcount <= 0:
        return []

    results = []
    for row in response:
        logger.debug(f"actions(): found {row['message_id']}")

        amount = row["amount"]
        if amount is not None:
            amount = amount.normalize()

        if row["path"] is not None:
            message = nyantip.reddit.comment(row["message_id"])
        else:
            message = nyantip.reddit.inbox.message(row["message_id"])

        if message.author is None:
            logger.warning("Cannot process item missing author. %r", message)
            continue

        results.append(
            Action(
                action=action,
                amount=amount,
                destination=row["destination"],
                message=message,
                nyantip=nyantip,
            )
        )

    return results


def check_action(**kwargs):
    return actions(**kwargs, _check=True)



import logging

from praw.exceptions import RedditAPIException
from praw.models import Comment
from prawcore.exceptions import Forbidden, NotFound

from .util import log_function

logger = logging.getLogger(__package__)


class User(object):
    def __init__(self, *, name, nyantip, redditor=None):
        assert isinstance(name, str)
        self.name = name
        self.nyantip = nyantip
        self.redditor = redditor

    def __eq__(self, other):
        return isinstance(other, User) and self.name.lower() == other.name.lower()

    def __repr__(self):
        return f"<User name={self.name}>"

    def __str__(self):
        return self.name

    def balance(self, *, kind):
        return self.nyantip.coin.balance(
            minconf=self.nyantip.config["coin"]["minconf"][kind], user=self.name
        )

    def is_redditor(self):
        if self.redditor is not None:
            return bool(self.redditor)
        self.redditor = self.nyantip.reddit.redditor(self.name)

        try:
            self.redditor.created_utc
        except NotFound:
            self.redditor = False

        return bool(self.redditor)

    def is_registered(self):
        return bool(
            self.nyantip.database.execute(
                "SELECT 1 FROM users WHERE username=%s", self
            ).one_or_none()
        )

    @log_function(klass="User")
    def message(self, *, body, message=None, reply_to_comment=False, subject):
        assert self.redditor is not None

        if message and (
            reply_to_comment
            or not (isinstance(message, Comment) or message.was_comment)
        ):
            assert self.redditor == message.author
            logger.debug(f"({self.redditor}): replying to message {message.id}")
            try:
                message.reply(body)
                return
            except RedditAPIException as exception:
                was_deleted = False
                for subexception in exception.items:
                    if subexception.error_type == "DELETED_COMMENT":
                        was_deleted = True
                        logger.debug(f"({self.redditor}): comment was deleted")
                if not was_deleted:
                    raise
            except Forbidden:
                logger.warning(f"Could not reply to message {message.id}")

        logger.debug(f"({self.name}): sending message {subject}")
        self.redditor.message(message=body, subject=subject)

    @log_function(klass="User")
    def register(self):
        address = self.nyantip.coin.generate_address(user=self.name)
        logger.info(f"register({self.name}): got {self.nyantip.coin} address {address}")
        self.nyantip.database.execute(
            "INSERT INTO users (address,username) VALUES (%s, %s)", (address, self)
        )



"""nyantip setup.py"""

import re
from codecs import open
from os import path

from setuptools import find_packages, setup

PACKAGE_NAME = "nyantip"
HERE = path.abspath(path.dirname(__file__))
with open(path.join(HERE, "README.md"), encoding="utf-8") as fp:
    README = fp.read()
with open(path.join(HERE, PACKAGE_NAME, "const.py"), encoding="utf-8") as fp:
    VERSION = re.search('__version__ = "([^"]+)"', fp.read()).group(1)

extras = {
    "gpg": ["python-gnupg == 0.4.7"],
    "lint": [
        "black",
        "flake8",
        "isort",
    ],
}

setup(
    author="Bryce Boe",
    author_email="bbzbryce@gmail.com",
    description="Nyancoin tip bot for Reddit.",
    entry_points={"console_scripts": [f"{PACKAGE_NAME}={PACKAGE_NAME}:main"]},
    extras_require=extras,
    include_package_data=True,
    install_requires=[
        "Jinja2 ~= 3.0",
        "PyYAML ~= 6.0",
        "mysqlclient ~= 2.2",
        "praw ~= 7.3",
        "python-bitcoinrpc == 1.0",  # Lock version incase someone does something malicous to this package.
        "sqlalchemy ~= 1.4",
    ],
    license="GPL",
    long_description=README,
    name=PACKAGE_NAME,
    package_data={"": ["LICENSE"]},
    packages=find_packages(),
    project_urls={"Source Code": "https://github.com/bboe/nyantip"},
    python_requires="~= 3.6",
    version=VERSION,
)



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
    This file is part of ALTcointip.

    ALTcointip is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    ALTcointip is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ALTcointip.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from datetime import datetime
from decimal import Decimal
from urllib.parse import quote_plus

from prawcore.exceptions import NotFound

MAX_WIKI_CONTENT = 511950  # Bytes

logger = logging.getLogger(__package__)


def format_coin(config, quantity):
    return f"{quantity:f} {config['coin']['symbol']}"


def format_value(*, compact=False, config, key, username=None, value):
    if not value:
        return "-"
    if isinstance(value, Decimal):
        return format_coin(config, value.normalize())
    if isinstance(value, datetime):
        return value.isoformat(" ", "minutes")
    if key == "comment":
        return f"[link]({value})"
    if key == "destination" and len(value) > 20:
        return f"[{value[:6]}...{value[-5:]}]({config['coin']['explorer']['address']}{value})"
    if key == "status":
        return "✓" if value == "completed" else value
    if key in ("destination", "source"):
        is_username = username and value.lower() == username.lower()
        user_string = f"**u/{value}**" if is_username else f"u/{value}"
        if compact:
            return user_string
        if not is_username:
            user_string += (
                f"^[[stats]](/r/{config['reddit']['subreddit']}/wiki/stats_{value})"
            )
        return user_string
    return value


def update_stats(nyantip=None):
    lines = []

    for stat, config in sorted(nyantip.config["sql"]["globalstats"].items()):
        logger.debug(f"update_stats(): getting stats for '{stat}'")

        total = nyantip.database.execute(config["query"]).scalar_one()

        lines.append(f"\n\n### {config['name']}\n")
        lines.append(f"{config['description']}: **{total}**\n")

    update_wiki(lines=lines, nyantip=nyantip, page="stats")


def update_tips(nyantip=None):
    tips = [f"### {nyantip.config['coin']['name']} Completed Tips\n"]

    result = nyantip.database.execute(nyantip.config["sql"]["tips"])
    tips.append("|".join(result.keys()))
    tips.append("|".join([":---"] * len(result.keys())))

    for row in result:
        values = []
        for key in result.keys():
            values.append(format_value(config=nyantip.config, key=key, value=row[key]))
        tips.append("|".join(values))

    update_wiki(lines=tips, nyantip=nyantip, page="tips")


def update_user_stats(*, nyantip, username):
    user_stats = [f"### Tipping Summary for u/{username}\n"]

    total = nyantip.database.execute(
        nyantip.config["sql"]["userstats"]["total_tipped"], username
    ).scalar_one_or_none()
    if total:
        user_stats.append(
            f"Total Tipped: {format_coin(nyantip.config, total.normalize())}\n"
        )

    total = nyantip.database.execute(
        nyantip.config["sql"]["userstats"]["total_received"], username
    ).scalar_one_or_none()
    if total:
        user_stats.append(
            f"Total Received: {format_coin(nyantip.config, total.normalize())}\n"
        )

    user_stats.append("#### History\n")
    result = nyantip.database.execute(
        nyantip.config["sql"]["userstats"]["history"], (username, username)
    )
    if result.rowcount <= 0:
        logger.debug(f"update_user_stats(): skipping {username} with no history")
        return

    user_stats.append("|".join(result.keys()))
    user_stats.append("|".join([":---"] * len(result.keys())))

    for row in result:
        history_entry = []
        for key in result.keys():
            history_entry.append(
                format_value(
                    config=nyantip.config,
                    key=key,
                    username=username,
                    value=row[key],
                )
            )
        user_stats.append("|".join(history_entry))

    update_wiki(
        lines=user_stats,
        nyantip=nyantip,
        page=f"stats_{username}",
    )


def update_wiki(*, lines, nyantip, page):
    subreddit = nyantip.config["reddit"]["subreddit"]
    content = wiki_fit(lines=lines)
    wiki = nyantip.reddit.subreddit(subreddit).wiki[page]
    try:
        previous_content = wiki.content_md.strip()
    except NotFound:
        previous_content = None
    if content.strip() != previous_content:
        logger.debug(f"update_user_stats(): updating wiki {subreddit}/{page}")
        wiki.edit(content=content)
    else:
        logger.debug(
            f"update_user_stats(): content not changed on wiki {subreddit}/{page}"
        )


def wiki_fit(*, lines):
    content = None
    end = size = len(lines)
    for i in range(end.bit_length() + 1):
        content = "\n".join(lines[:end])
        size = max(1, round(size / 2))
        if len(quote_plus(content)) <= MAX_WIKI_CONTENT:
            if (
                end == len(lines)
                or len(quote_plus("\n".join(lines[: end + 1]))) > MAX_WIKI_CONTENT
            ):
                return content
            end += size
        else:
            end -= size
    raise Exception(f"wiki_fit didn't terminate in {i} iterations")



from bitcoinrpc.authproxy import AuthServiceProxy


def close_connection(connection, function_name):
    function = getattr(connection, function_name)

    def wrapped(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        finally:
            connection._AuthServiceProxy__conn.close()

    return wrapped


class Rpc:
    def __init__(self, url):
        self._connection = AuthServiceProxy(url)

    def __getattr__(self, attribute):
        return close_connection(connection=self._connection, function_name=attribute)
